"""Models - Playlist"""

from rkstreamer.models.data import Playlist, PlaylistSearch, Song
from rkstreamer.models.song import JioSaavnSongQueue
from rkstreamer.interfaces.models import IPlaylistModel
from rkstreamer.services.playlist import JioSaavnPlaylistProvider
from rkstreamer.types import (
    PlaylistType,
    PlaylistSearchType,
    PlaylistSearchIndexType,
    NetworkProviderType
)
from rkstreamer.models.exceptions import InvalidInput


class JioSaavnPlaylistModel(IPlaylistModel):
    """Playlist model implemented for Jio Saavn service"""

    def __init__(self, network_provider: NetworkProviderType) -> None:
        self.stream_provider = JioSaavnPlaylistProvider(
            client=network_provider)
        self.queue = JioSaavnSongQueue()
        self.indexed_playlists = {}
        self.playlist_songs = []

    def _create_playlist(self, **kwargs) -> PlaylistType:
        return Playlist(**kwargs)

    def _create_search_playlist(self, **kwargs) -> PlaylistSearchType:
        return PlaylistSearch(**kwargs)

    def _create_search_playlist_index(self, playlists) -> PlaylistSearchIndexType:
        self.indexed_playlists = {count: self._create_search_playlist(**playlist)
                                  for count, playlist in enumerate(playlists, 1)}
        return self.indexed_playlists

    def search(self, search_string: str, **kwargs) -> PlaylistSearchIndexType:
        response = self.stream_provider.search_playlists(
            search_string, **kwargs)
        return self._create_search_playlist_index(response)

    def select(self, selection: int, **kwargs) -> PlaylistType:
        selected_plist = self.indexed_playlists.get(int(selection))
        if selected_plist:
            plist_songs = self.stream_provider.select_playlist(
                selected_plist.token, **kwargs)
            if plist_songs:
                selected_plist.__dict__.update(
                    {'songs': [Song(**song, status='Loaded') for song in plist_songs]})
                return self._create_playlist(**selected_plist.__dict__)
        raise InvalidInput("Invalid playlist selection input provided.")



"""
DI module for network requests
"""
from __future__ import annotations
from typing import TYPE_CHECKING
from ._interface import ABC, abstractmethod

if TYPE_CHECKING:
    from rkstreamer.types import (
        NetworkProviderResponseType
    )

class INetworkProvider(ABC):
    """Interface for Network Providers"""

    @abstractmethod
    def get(self, **kwargs) -> NetworkProviderResponseType:
        """GET request method"""


class INetworkProviderResponse(ABC):
    """Interface for Response object returned by Network Providers"""

    @abstractmethod
    def json(self) -> dict:
        """Return JSONified response"""

    @property
    @abstractmethod
    def headers(self) -> dict:
        """Response headers"""

    @property
    @abstractmethod
    def status_code(self) -> int:
        """Response code"""

    @abstractmethod
    def raise_for_status(self) -> None:
        """Raise for validating response status"""



"""
DI module for domain models
"""

from ._interface import ABC, abstractmethod


class IModel(ABC):
    """Interface for model"""

    @abstractmethod
    def search(self, search_string: str, **kwargs):
        """Search item"""

    @abstractmethod
    def select(self, selection: int, **kwargs):
        """Select item"""


class IQueue(ABC):
    """Interface for queue"""

    @abstractmethod
    def add(self, entity):
        """Add entity to queue"""

    @abstractmethod
    def remove(self, index: int):
        """Remove entity from queue and updates the index"""

    @abstractmethod
    def fetch(self, index: int):
        """fetch entity from queue index"""

    @property
    @abstractmethod
    def get_queue(self):
        """Returns the queue"""

    @property
    @abstractmethod
    def get_indexed_queue(self):
        """Returns the indexed queue"""


class ISongModel(IModel):
    """Interface for Song model"""

    @abstractmethod
    def get_song_url(self, data):
        """Get Song stream url"""

    @abstractmethod
    def get_related_songs(self, data):
        """Load related songs"""


class ISongQueue(IQueue):
    """Inferace for Song queue"""

    @abstractmethod
    def add_related_songs(self, songs):
        """Add related songs to queue"""

    @abstractmethod
    def update_qstatus(self, status, stream_url):
        """Update the main queue status"""

    @abstractmethod
    def pop_rsong(self):
        """Get related song from Rqueue list"""

    @abstractmethod
    def get_rsong_index(self, index: int):
        """Get related song by index"""

    @abstractmethod
    def remove_rsong_index(self, index: int):
        """Remove related song by index"""

    @abstractmethod
    def update_rqueue(self, rsongs):
        """Updates the related song queue"""

    @property
    @abstractmethod
    def get_rsongs(self):
        """Get rsongs list"""


class IAlbumModel(IModel):
    """Inferface for Album model"""

    @abstractmethod
    def select_song_from_album(self, selection: int):
        """Select song from album"""


class IPlaylistModel(IModel):
    """Interface for Playlist model"""



"""
DI modules for Player
"""

from ._interface import ABC, abstractmethod

class MusicPlayer(ABC):
    """Interface for music player"""

    @abstractmethod
    def add_media(self, media_url: str):
        """Add media to queue"""

    @abstractmethod
    def remove_media(self, media_url: str):
        """Add media to queue"""

    @abstractmethod
    def play_media(self, media_url: str):
        """Play media"""

class MusicPlayerControls(ABC):
    """Interface for music player cntrols"""

    @abstractmethod
    def player_controls(self, user_input: str):
        """Player controls"""



"""Tests: Search playlist provider"""

from rkstreamer.services.network import PyRequests
from rkstreamer.services.playlist import JioSaavnPlaylistProvider


proxy = {'https': 'http://127.0.0.1:8888'}
test = JioSaavnPlaylistProvider(client=PyRequests(verify=False, proxy=proxy))
search = test.search_playlists('akon')
print(search)



"""Test: Get album using ID"""

from rkstreamer.services.network import PyRequests
from rkstreamer.services.album import JioSaavnAlbumProvider


test = JioSaavnAlbumProvider(client=PyRequests(verify=False))
search = test.select_album_id('qKErkhPpdTE_')
print(search)


"""
RK Streamer

A command-line music player that allows you to play/stream songs, albums, playlists.

Author: Ray A.
Email: ray@raysecure.ml

(c) 2023 Ray A. All rights reserved.

License:
This project is licensed under the terms of the MIT License.
See the LICENSE file for more information.
"""

from rkstreamer.state import State, StateMachine
from rkstreamer.services.player import PyVLCPlayer
from rkstreamer.services.network import PyRequests
from rkstreamer.models import (
    JioSaavnSongModel,
    JioSaavnAlbumModel,
    JioSaavnPlaylistModel
)
from rkstreamer.views import (
    JioSaavnSongView,
    JioSaavnAlbumView,
    JioSaavnPlaylistView
)
from rkstreamer.controllers import (
    JioSaavnSongController,
    JioSaavnAlbumController,
    JioSaavnPlaylistController
)

# proxy = {'https': 'http://127.0.0.1:8888'}
pyrequests = PyRequests(proxy=None)

song_controller = JioSaavnSongController(
    model=JioSaavnSongModel(pyrequests),
    view=JioSaavnSongView(player=PyVLCPlayer())
)

album_controller = JioSaavnAlbumController(
    model=JioSaavnAlbumModel(pyrequests),
    view=JioSaavnAlbumView(player=PyVLCPlayer())
)

plist_controller = JioSaavnPlaylistController(
    model=JioSaavnPlaylistModel(pyrequests),
    view=JioSaavnPlaylistView(player=PyVLCPlayer())
)

song = State(
    "song",
    "Enter the song name: ",
    song_controller
)

album = State(
    "album",
    "Enter the album name: ",
    album_controller
)

playlist = State(
    "plist",
    "Enter the playlist name: ",
    plist_controller
)

print(r"""

*** RK Streamer ***

@Powered by Jio Saavn

""")


streamer = StateMachine()
streamer.add_state({'song': song, 'album': album, 'plist': playlist})
streamer.set_start_state('song')
streamer.trigger()



"""DI module for Database API
"""

from ._interface import ABC, abstractmethod


class IDatabaseProvider(ABC):
    """Interface for Database Providers"""

    @abstractmethod
    def connect(self):
        """Connect to DB"""

    @abstractmethod
    def check(self, name: str):
        """Checking DB content"""

    @abstractmethod
    def read(self, name: str):
        """Reading DB content"""

    @abstractmethod
    def write(self, name: str):
        """Writing DB content"""

    @abstractmethod
    def close(self):
        """Close DB connection"""



"""Models - Exception module"""

class InvalidInput(Exception):
    """Invalid/unexpected input provided"""

class QueueException(Exception):
    """Base exception for Queue models"""

class AddMediaError(QueueException):
    """Add media to queue failed"""

class RemoveMediaError(QueueException):
    """Remove media from queue failed"""

class GetMediaError(QueueException):
    """Get media from queue failed"""

class MediaNotFound(QueueException):
    """Media not found in queue"""



"""RKStreamer - State and State machine module"""

from rkstreamer.models.exceptions import InvalidInput, QueueException

class State:
    """State class"""

    def __init__(self, name, prompt, controller):
        self.name = name
        self.prompt = prompt
        self.controller = controller

    def handle_input(self, user_input):
        """Handle user input"""
        self.controller.handle_input(user_input)


class StateMachine:
    """State machine"""

    def __init__(self):
        self.states = {}
        self.current_state = None

    def add_state(self, state):
        """Add state to states list"""
        if isinstance(state, dict):
            self.states = state
        else:
            self.states[state.name] = state

    def set_start_state(self, state_name):
        """Setting the start date - init"""
        self.current_state = self.states[state_name]

    def trigger(self):
        """Triggers the state change or handle current state"""
        while True:
            try:
                user_input = input(self.current_state.prompt)
            except KeyboardInterrupt:
                raise SystemExit("\n\nTata! See you soon!!") from None
            if user_input.lower().startswith('-e'):
                raise SystemExit('Tata!')
            if user_input.startswith("--"):
                state_name = user_input[2:]
                if (state_name != self.current_state.name) and (state_name in self.states):
                    # stop player when switching mode.
                    self.current_state.controller.view.stop()
                    self.current_state = self.states[state_name]
            else:
                if user_input:
                    try:
                        self.current_state.handle_input(user_input)
                    except InvalidInput as exc:
                        print(f"Error: {exc.__class__.__name__}, Desc: 'Invalid Input'")
                    except QueueException as exc:
                        print(f"Error: {exc.__class__.__name__}, Desc: 'Invalid Queue operation'")



"""
DI Module for stream providers API
"""

from ._interface import ABC,abstractmethod

class ISongProvider(ABC):
    """Interface for song provider"""

    @abstractmethod
    def search_songs(self, search_string: str, **kwargs):
        """Search songs"""

    @abstractmethod
    def select_song(self, arg: str, **kwargs):
        """Select song"""


class IAlbumProvider(ABC):
    """Interface for album provider"""

    @abstractmethod
    def search_albums(self, search_string: str, **kwargs):
        """Search albums"""

    @abstractmethod
    def select_album(self, arg: str, **kwargs):
        """Select album"""


class IPlaylistProvider(ABC):
    """Interface for playlist provider"""

    @abstractmethod
    def search_playlists(self, search_string: str, **kwargs):
        """Search playlists"""

    @abstractmethod
    def select_playlist(self, arg: int, **kwargs):
        """Select playlist"""



"""
Album view implementation module
"""

from rkstreamer.types import MusicPlayerControlsType
from rkstreamer.views.song import JioSaavnSongView


class JioSaavnAlbumView(JioSaavnSongView):
    """Jio Saavn Album view"""

    def __init__(self, player: MusicPlayerControlsType) -> None:
        self.player = player
        super().__init__(player)

    def display(self, entity) -> None:
        print('\n\033[33m\033[01m***Albums Search***\033[0m\n')
        for count, album in entity.items():
            print(f"{count}- {album.name} ({album.language})")
            print(f"Music - {album.artists}")
            print(f"Song Count - {album.song_count}")
            print()

    def display_album_songs(self, entity) -> None:
        """Display album songs"""
        print('\n\033[33m\033[01m***Albums Songs***\033[0m\n')
        for count, song in entity.songs.items():
            print(f"A{count}- {song.name}")
            print(f"Artists - {song.artists}")
            print()



"""Test: Search albums"""

from rkstreamer.services.network import PyRequests
from rkstreamer.services.album import JioSaavnAlbumProvider


test = JioSaavnAlbumProvider(client=PyRequests(verify=False))
search = test.search_albums('madhrasapattinam')
assert search[0].items()
print(search[0].items())


"""Types"""

from typing import NewType
from rkstreamer.interfaces import (
    Command,
    INetworkProvider,
    INetworkProviderResponse,
    MusicPlayer,
    MusicPlayerControls)

NetworkProviderType = NewType('INetworkProvider', INetworkProvider)
NetworkProviderResponseType = NewType('INetworkProviderResponse', INetworkProviderResponse)
CommandType = NewType('Command', Command)
MusicPlayerType = NewType('MusicPlayer', MusicPlayer)
MusicPlayerControlsType = NewType('MusicPlayerControls', MusicPlayerControls)


"""
Models - Song
"""

from typing import Optional
from rkstreamer.models.data import Song, SongSearch, SongQueue
from rkstreamer.interfaces.models import ISongModel, ISongQueue
from rkstreamer.services.song import JioSaavnSongProvider
from rkstreamer.types import (
    SongListRawType,
    SongListType,
    SongIndexType,
    SongType,
    SongSearchType,
    SongSearchIndexType,
    SongQueueType,
    SongQueueIndexType,
    SongQueueModelType,
    NetworkProviderType)
from rkstreamer.models.exceptions import (
    AddMediaError,
    RemoveMediaError,
    GetMediaError,
    MediaNotFound,
    InvalidInput
)


class JioSaavnSongModel(ISongModel):
    """Song model implemented for Jio Saavn Service"""

    def __init__(self, network_provider: NetworkProviderType) -> None:
        self.network_provider = network_provider
        self.stream_provider = JioSaavnSongProvider(
            client=self.network_provider)
        self.queue: SongQueueModelType = JioSaavnSongQueue()
        self.indexed_search_songs: SongSearchIndexType = {}
        self._recomm_song_index = 1

    def _create_song(self, **kwargs) -> SongType:
        return Song(**kwargs)

    def _create_search_song(self, **kwargs) -> SongSearchType:
        return SongSearch(**kwargs)

    def _create_recomm_song(self, songs: SongListRawType) -> SongIndexType:
        recomm_songs = {}
        _songs: SongListType = [self._create_song(**song) for song in songs]
        for song in _songs:
            if song not in self.queue.rsongs_copy_:
                recomm_songs.update({self._recomm_song_index: song})
                self._recomm_song_index += 1
        return recomm_songs

    def _create_search_song_index(self, songs: SongListRawType) -> SongSearchIndexType:
        self.indexed_search_songs = {count: self._create_search_song(**song)
                                     for count, song in enumerate(songs, 1)}
        return self.indexed_search_songs

    def search(self, search_string: str, **kwargs) -> SongSearchIndexType:
        """Search Song using the search string and creates a Song Search data model"""
        search_result: SongListRawType = self.stream_provider.search_songs(
            search_string, **kwargs)
        return self._create_search_song_index(search_result)

    def select(self, selection: int, **kwargs) -> Optional[SongType]:
        """Select Song from the Song Search Index and returns the selected Song"""
        selected_song = self.indexed_search_songs.get(int(selection))
        if selected_song:
            song_url = self.stream_provider.select_song(
                selected_song.token, **kwargs)
            if song_url:
                selected_song.__dict__.update({'stream_url': song_url})
                selected_song.__dict__.update({'status': 'Loaded'})
                return self._create_song(**selected_song.__dict__)
        raise InvalidInput("Invalid song selection input provided.")

    def get_song_url(self, data: str) -> str:
        """Get the song's stream url using Enc Url Token - used for rsongs download"""
        stream_url = self.stream_provider.select_song(data)
        return stream_url

    def get_related_songs(self, data: str) -> SongIndexType:
        """Gets recommended songs using song_id and updates the RQueue"""
        recomm_songs_raw: SongListRawType = self.stream_provider.get_recomm_songs(
            data)
        if recomm_songs_raw:
            return self._create_recomm_song(recomm_songs_raw)


class JioSaavnSongQueue(ISongQueue):
    """Song queue implemented for Jio Saavn"""

    def __init__(self) -> None:
        self.queue: SongQueueType = SongQueue([])
        self._indexed_queue = {}
        self.rsongs_copy_ = set()
        self.rsongs_list = {}
        self.current_playing_song = None

    def _create_song(self, **kwargs) -> SongType:
        return Song(**kwargs)

    def _check_media(self, song: SongType) -> bool:
        """Checks the presence of media"""
        if song in self.queue.songs:
            return True
        return False

    def _normalize_queue_index(self) -> None:
        """Updates the queue with index"""
        self._indexed_queue = dict(enumerate(self.queue.songs, 1))

    def add_playlist(self, entity: SongType) -> None:
        """Add playlist songs"""
        for song in entity.songs:
            if not self._check_media(song):
                self.queue.songs.append(song)
                self._normalize_queue_index()
        print(f"\n\033[01m\033[32mAdded: '{entity.name}'\033[0m")
        # print()

    def change_loaded_status(self) -> None:
        """Change loaded status for all songs with 'loaded' status"""
        for song in self.queue.songs:
            if song.status == 'Loaded':
                song.status = '\033[31mPlayed\033[0m'

    def change_loaded_status_before(self, entity: SongType) -> None:
        """Change loaded status for songs that before the called one."""
        for song in self.queue.songs:
            if entity == song:
                break  # loop until this song is found.
            if song.status == 'Loaded':
                song.status = '\033[31mPlayed\033[0m'
                # change the status for all songs thats above this one.

    def flush_queue(self) -> None:
        """Flushes the queue"""
        self.queue.songs.clear()
        self._normalize_queue_index()

    def add(self, entity: SongType) -> None:
        """Add Song to Queue"""
        if not self._check_media(entity):
            self.queue.songs.append(entity)
            self._normalize_queue_index()
            print(f"\n\033[01m\033[32mAdded: '{entity.name}'\033[0m")
            # print()
        if not self._check_media(entity):
            raise AddMediaError("Failed to add media to queue")
        if entity not in self._indexed_queue.values():
            raise AddMediaError("Failed to add media to queue index")

    def remove(self, index: int) -> SongListType:
        """Remove Songs by Index value"""
        removed_songs = []
        for value in list(index):
            song = self.fetch(value)
            if self._check_media(song):
                self.queue.songs.remove(song)
                self._indexed_queue.pop(int(value))
                removed_songs.append(song)
                print(f"\n\033[93mRemoved: '{song.name}'\033[0m")
                # print()
            else:
                raise MediaNotFound("Media not found in queue to delete")
            if song in self._indexed_queue.values():
                raise RemoveMediaError(
                    "Failed to remove media from queue index")
        self._normalize_queue_index()
        return removed_songs

    def fetch(self, index: int) -> SongType:
        """Fetch the song from Queue Index"""
        song = self._indexed_queue.get(int(index))
        if not song:
            raise GetMediaError("Failed to get media from queue index")
        return song

    def add_related_songs(self, songs: SongListRawType) -> None:
        """Add recommended songs to queue index"""
        queue_length = len(self.queue.songs)
        for song in songs:
            if not self._check_media(song):
                self.queue.songs.append(song)
        for count, song in enumerate(self.queue.songs, queue_length+1):
            self._indexed_queue.update({count: song})

    def update_qstatus(self, status: str, stream_url: str) -> Optional[SongType]:
        """Update 'played' status for songs in Queue and returns Song object if it's successful.
        Also, Updates the current_playing_song attr."""
        for song in self.queue.songs:
            if stream_url == song.stream_url:
                self.current_playing_song = song
                if song.status == 'Loaded':
                    print(f"\n\033[31m>>> Playing '{song.name}' <<<\033[0m")
                    song.status = status
                    return song
        return None

    @property
    def get_queue(self) -> SongQueueType:
        """returns song queue"""
        return self.queue

    @property
    def get_indexed_queue(self) -> SongQueueIndexType:
        """returns the indexed queue"""
        return self._indexed_queue

    def check_status(self, queue: SongQueueType) -> bool:
        """Checks the queue status for 'played' songs"""
        for song in queue.songs:
            if song.status == '\033[31mPlayed\033[0m':
                continue
            else:
                return False
        return True

    def pop_rsong(self) -> Optional[SongType]:
        """Pop rsong from its queue and move it to main queue.
        Change the song status to 'Loaded'"""
        if self.rsongs_list:
            first_rsong = next(iter(self.rsongs_list))
            rsong = self.rsongs_list.pop(first_rsong)
            # rsong = self.rsongs_list.pop(0)
            rsong.status = 'Loaded'
            return rsong
        return None

    def get_rsong_index(self, index: int) -> Optional[SongType]:
        """Get RS song from the given index"""
        try:
            rsong = self.rsongs_list.pop(index)
            return rsong
        except (IndexError,KeyError):
            raise GetMediaError("Failed to get song from RS Queue") from None
        return None

    def remove_rsong_index(self, index: int) -> Optional[SongType]:
        """Removes the RS song from its queue"""
        try:
            rsong = self.rsongs_list.pop(index)
            if rsong:
                print(f"\n\033[33mRemoved: '{rsong.name}'\033[0m")
        except (IndexError,KeyError):
            raise RemoveMediaError("Failed to remove song from RS Queue") from None

    @property
    def get_rsongs(self) -> SongIndexType:
        """Get the rsongs list"""
        return self.rsongs_list

    def update_rqueue(self, rsongs: SongIndexType):
        """Update the rsongs list"""
        self.rsongs_list.update(rsongs)
        self.rsongs_copy_.update(rsongs.values())



"""Playlist controller implemenation"""

import re
from typing import Union
from rkstreamer.interfaces.patterns import Command
from rkstreamer.interfaces.controllers import IController
from rkstreamer.controllers.enums import ControllerEnum
from rkstreamer.controllers.patterns import PlayerControlsCommand
from rkstreamer.controllers.queue import SongQueueCommand
from rkstreamer.utils.helper import parse_input, PLIST_PATTERN
from rkstreamer.types import (
    PlaylistModelType,
    PlaylistViewType,
    PlaylistControllerType,
    CommandType,
    SongType
)


class JioSaavnPlaylistController(IController):
    """Playlist Controller implemented for Jio Saavn model"""

    def __init__(self, model: PlaylistModelType, view: PlaylistViewType) -> None:
        self.model = model
        self.view = view
        self.commands = {
            ControllerEnum.QUEUE: SongQueueCommand(self),
            ControllerEnum.CONTROLS: PlayerControlsCommand(self),
            ControllerEnum.PVIEW: PlaylistViewCommand(self),
            str: PlaylistSearchCommand(self),
            int: PlaylistSelectCommand(self),
        }
        self.view.set_controller_callback(self.uow_update_song_status)

    def handle_input(self, user_input: Union[str, int]):
        if user_input.startswith('-'):
            re_match = re.match(r'(-\w{1})', user_input)
            try:
                enum_obj = ControllerEnum(re_match.group(1))
                command: CommandType = self.commands.get(enum_obj)
                command.execute(user_input)
            except (ValueError, AttributeError) as exc:
                print(f"{exc.__class__, 'Invalid input. Please try again'}")
        else:
            input_type = type(parse_input(user_input))
            command: CommandType = self.commands.get(input_type)
            command.execute(user_input)

    def uow_update_song_status(self, status: str, stream_url: str):
        """)override) Updating song status in queue to "Played"""
        return self.model.queue.update_qstatus(status, stream_url)

    def uow_add_songs_queue(self, playlist):
        """UOW: (override) Add songs to queue & don't play it"""
        self.model.queue.add_playlist(playlist)
        songs = []
        for song in playlist.songs:
            songs.append(song.stream_url)
        self.view.add_media_list(songs)

    def uow_play_songs(self, playlist):
        """UOW: (override) play the songs"""
        self.view.stop()
        self.model.queue.flush_queue()  # clears the queue before setting the new plist
        self.uow_add_songs_queue(playlist)
        self.view.play()

    def uow_play_songs_remove_loaded_before(self, song: SongType):
        """UOW: plays the song from song queue - playlist mode so don't change the played status
        of other songs.
        
        name of the function has to be kept as it is
        because it's being called by SongQueueCommand."""
        self.view.play_media(song)


class PlaylistSelectCommand(Command):
    """Playlist Select command implementation"""

    def __init__(self, controller: PlaylistControllerType):
        self.controller = controller
        self.model: PlaylistModelType = self.controller.model
        self.view: PlaylistViewType = self.controller.view

    def execute(self, user_input: int):
        playlist = self.model.select(user_input)
        self.controller.uow_play_songs(playlist)


class PlaylistSearchCommand(Command):
    """Album Search"""

    def __init__(self, controller: PlaylistControllerType):
        self.controller = controller
        self.model: PlaylistModelType = self.controller.model
        self.view: PlaylistViewType = self.controller.view

    def execute(self, user_input: str):
        match_input = re.match(PLIST_PATTERN, user_input)
        if match_input:
            plist = match_input.group('plist').strip()
            match_input.groupdict().pop('plist')
            search_results = self.model.search(plist, **match_input.groupdict())
            self.view.display(search_results)
        else:
            print("Invalid Input provided!")
        # search_results = self.model.search(user_input)
        # self.view.display(search_results)


class PlaylistViewCommand(Command):
    """Album Search"""

    def __init__(self, controller: PlaylistControllerType):
        self.controller = controller
        self.model: PlaylistModelType = self.controller.model
        self.view: PlaylistViewType = self.controller.view

    def execute(self, user_input: str):
        user_input = user_input.strip('-v')
        search_results = self.model.select(user_input, view=True)
        self.view.display_playlist_songs(search_results)



"""
DI module for Controllers
"""

from typing import Union
from ._interface import ABC, abstractmethod


class IController(ABC):
    """Interface for Song controller"""

    @abstractmethod
    def handle_input(self, user_input: Union[str, int]) -> None:
        """Process user input"""


class ISongController(IController):
    """Interface for Song controller"""

    @abstractmethod
    def monitor_queue_pull_rsong(self):
        """Pull rsong by monitoring queue status"""

    @abstractmethod
    def uow_update_song_status(self, status: str, stream_url: str):
        """UOW: updates song status in queue"""

    @abstractmethod
    def uow_add_rsongs_rqueue(self, data: str):
        """UOW: Add rsongs to rqueue"""

    @abstractmethod
    def uow_add_songs_queue(self, song):
        """Add songs to queue"""

    @abstractmethod
    def uow_play_songs_remove_loaded(self, song):
        """Sets all loaded songs to 'Played' when playing songs
        from search or rsongs queue"""


class IAlbumController(ISongController):
    """Interface for Album controller"""


class IPlaylistController(IController):
    """Interface for Playlist controller"""

    @abstractmethod
    def uow_add_songs_queue(self, song):
        """Add songs to queue"""

    @abstractmethod
    def uow_play_songs(self, song):
        """sets media list and play"""



"""Module with helper functions & classes"""


def parse_input(input_str: str):
    """Return input based on the type"""
    if input_str.isdigit():
        return int(input_str)
    return input_str


LANGUAGES = ['tamil', 'english', 'hindi', 'telugu',
             'kannada', 'spanish', 'latin', 'malayalam',
             'bhojpuri', 'urdu', 'punjabi', 'gujarati',
             'haryanvi', 'bengali', 'rajasthani', 'assamese',
             'marathi']

SPACE = r"([\s]+)?"
SONG = r"(?P<song>([\w\s]+))"
ALBUM = r"(?P<album>([\w\s]+))"
PLIST = r"(?P<plist>([\w\s]+))"
LANG = r"(\-[l]:(?P<lang>\w+){1,})?"
NUM = r"(\-[n]:(?P<num>\d{1,}))?"
BITRATE = r"(\-[b]:(?P<bitrate>(160|320)))?"
RSONGS = r"(\-[r]:(?P<rsongs>\d{1,2}))?"

SONG_PATTERN = SONG+SPACE+LANG+SPACE+NUM+SPACE+BITRATE+SPACE+RSONGS
ALBUM_PATTERN = ALBUM+SPACE+LANG+SPACE+NUM
PLIST_PATTERN = PLIST+SPACE+LANG+SPACE+NUM



"""Tests: Display album songs"""

from rkstreamer.services.player import PyVLCPlayer
from rkstreamer.controllers.album import JioSaavnAlbumController
from rkstreamer.models.album import JioSaavnAlbumModel
from rkstreamer.views.album import JioSaavnAlbumView
from rkstreamer.services.network import PyRequests

proxy = {'https': 'http://127.0.0.1:8888'}
pyrequests = PyRequests(proxy=None)

test = JioSaavnAlbumController(
    model=JioSaavnAlbumModel(pyrequests),
    view=JioSaavnAlbumView(player=PyVLCPlayer()))

while True:
    user_input = input("Enter the album name: ")
    if user_input:
        test.handle_input(user_input)



"""Test: Select song from album"""

from rkstreamer.services.network import PyRequests
from rkstreamer.models.album import JioSaavnAlbumModel


test = JioSaavnAlbumModel(network_provider=PyRequests(verify=False))
search = test.search('madhrasapattinam')
print(search)
select = test.select(1)
print(select)
song = test.select_song_from_album(1)
print(song)



"""Test: Select album"""

from rkstreamer.services.network import PyRequests
from rkstreamer.models.album import JioSaavnAlbumModel


test = JioSaavnAlbumModel(network_provider=PyRequests(verify=False))
search = test.search('madhrasapattinam')
print(search)


"""Data Models"""

from ._model import dataclass, field

@dataclass(unsafe_hash=True)
class SongBase():
    """Base for song model"""
    name: str = field(hash=True, compare=True)
    id: str = field(hash=False, compare=False, default=None, repr=False)
    artists: str = field(hash=False, compare=False, default=None)
    music: str = field(hash=False, compare=False, default=None)
    album_name: str = field(hash=False, compare=False, default=None)
    album_id: str = field(hash=False, compare=False, default=None, repr=False)
    duration: int = field(hash=False, compare=False, default=None, repr=False)


@dataclass(unsafe_hash=True)
class SongSearch(SongBase):
    """Song Search model'"""
    status: str = field(hash=False, compare=False, default='Fetched', repr=False)
    token: str = field(hash=False, compare=False, default=None, repr=False)
    language: str = field(hash=False, compare=False, default=None, repr=False)


@dataclass(unsafe_hash=True)
class Song(SongSearch):
    """Song model"""
    stream_url: str = field(hash=False, compare=False, default=None)


@dataclass
class SongQueue():
    """Song Queue model"""
    songs: list = field(default_factory=list)


@dataclass
class AlbumBase():
    """Album base model"""
    name: str = field(hash=True, compare=True)
    id: str = field(hash=False, compare=False, default=None, repr=False)
    artists: str = field(hash=False, compare=False, default=None)
    music: str = field(hash=False, compare=False, default=None)

@dataclass
class AlbumSearch(AlbumBase):
    """Album search model"""
    song_count: int = field(hash=False, default=None)
    language: str = field(hash=False, compare=False, default=None, repr=False)

@dataclass
class Album(AlbumSearch):
    """Album model"""
    songs: list = field(default_factory=list)


@dataclass
class PlaylistBase():
    """Playlist base model"""
    name: str = field(hash=True, compare=True)

@dataclass
class PlaylistSearch(PlaylistBase):
    """Playlist search model"""
    token: str = field(default=None, repr=False)
    song_count: int = field(default=None)
    language: str = field(hash=False, compare=False, default=None, repr=False)


@dataclass
class Playlist(PlaylistSearch):
    """Playlist model"""
    songs: dict = field(default_factory=dict)



"""Tests: Get related songs"""

from rkstreamer.services.network import PyRequests
from rkstreamer.models import JioSaavnSongModel

pyrequests = PyRequests()

song_model = JioSaavnSongModel(pyrequests)

recomm_songs = song_model.get_related_songs('elrx2wXJ')

print(recomm_songs)


"""
Playlist provider - API
"""

from __future__ import annotations
from typing import TYPE_CHECKING
from html import unescape
from urllib.parse import urlencode
from urllib3 import disable_warnings
from rkstreamer.interfaces.provider import IPlaylistProvider
from rkstreamer.utils.helper import LANGUAGES

if TYPE_CHECKING:
    from rkstreamer.types import (
        SongUrl,
        PListRawType,
        NetworkProviderType,
        NetworkProviderResponseType
    )

disable_warnings()  # Function to suppress the SSL Verification error.


class JioSaavnPlaylistProvider(IPlaylistProvider):
    """Jio Saavn - Playlist provider API"""

    API_BASE = "https://www.jiosaavn.com/api.php"

    PARAMS_DEFAULT = {'api_version': 4, '_format': 'json', '_marker': 0,
                      'ctx': 'web6dot0'}
    PARAMS_FTEXT = {'api_version': 4, '_format': 'text', '_marker': 0,
                    'ctx': 'web6dot0'}  # Format set to text for plist songs download.

    plist_search = {'__call': 'search.getPlaylistResults', 'p': 1, 'q': ''}

    plist_download = {'__call': 'webapi.get', 'token': '', 'type': 'playlist',
                      'includeMetaTags': 0, 'p': 1, 'n': 50}

    def __init__(self, client: NetworkProviderType) -> None:
        self.client = client

    def search_playlists(self, search_string: str, **kwargs) -> PListRawType:
        """Playlist search"""
        self.plist_search['q'] = search_string
        language = [kwargs.get('lang'),] \
            if kwargs.get('lang') \
            else LANGUAGES
        plist_request = self.client.get(
            url=f"{self.API_BASE}?{urlencode(self.plist_search|self.PARAMS_DEFAULT, safe='+')}")
        return self._parse_playlist(plist_request, lang=language)

    def _parse_playlist(self, response: NetworkProviderResponseType, **kwargs) -> PListRawType:
        """Parse playlist response"""
        return [{'name': plist['title'],
                'token': plist['perma_url'].split('/')[-1],
                'language': plist['more_info']['language'],
                 'song_count': plist['more_info']['song_count']}
                for plist in response.json()['results']
                if plist['more_info']['language'] in kwargs.get('lang')]

    def select_playlist(self, arg: str, **kwargs) -> PListRawType:
        self.plist_download['token'] = arg
        plist_select = self.client.get(
            url=self.API_BASE, params=self.plist_download | self.PARAMS_FTEXT)
        return self._parse_playlist_songs(plist_select, **kwargs)

    def _parse_playlist_songs(self, response: NetworkProviderResponseType, **kwargs) -> PListRawType:
        """Parse songs from playlist selection"""
        if kwargs.get('view'):
            return [{'name': unescape(song['song_for_player'])}
                    for song in response.json()['fullsongs']]

        return [{'name': unescape(song['song_for_player']),
                'stream_url': self._change_plist_song_url(song['download_url'])}
                for song in response.json()['fullsongs']]

    def _change_plist_song_url(self, song_url: str) -> SongUrl:
        """Change plist song urls with latest JS CDN host"""
        if 'SAR' in song_url:  # for handling URLs - https://h.saavncdn.com/*/SAR-*.mp3
            return song_url
        rm_pattern = song_url.removeprefix(
            'http://h.saavncdn.com').removeprefix('https://h.saavncdn.com').removesuffix('.mp3')
        new_pattern = "https://aac.saavncdn.com"+rm_pattern+"_96.mp4"
        return new_pattern



"""Tests: Search songs"""

from rkstreamer.services.network import PyRequests
from rkstreamer.services.song import JioSaavnSongProvider


test = JioSaavnSongProvider(client=PyRequests(verify=False))
search = test.search_songs('pokkal pokkum')
assert search[0].items()
print(search[0].items())



"""Tests: Display songs"""

from rkstreamer.services.player import PyVLCPlayer
from rkstreamer.controllers.song import JioSaavnSongController
from rkstreamer.models.song import JioSaavnSongModel
from rkstreamer.views.song import JioSaavnSongView
from rkstreamer.services.network import PyRequests

proxy = {'https': 'http://127.0.0.1:8888'}
pyrequests = PyRequests(proxy=None)

test = JioSaavnSongController(
    model=JioSaavnSongModel(pyrequests),
    view=JioSaavnSongView(player=PyVLCPlayer()))

while True:
    user_input = input("Enter the song name: ")
    if user_input:
        test.handle_input(user_input)



"""Album controller - module"""

import re
from typing import Union
from rkstreamer.interfaces.patterns import Command
from rkstreamer.controllers.enums import ControllerEnum
from rkstreamer.controllers.queue import AlbumQueueCommand, ReSongQueueCommand
from rkstreamer.controllers.patterns import (
    ControllerUtils,
    PlayerControlsCommand,
    GotoAlbumCommand
)
from rkstreamer.utils.helper import parse_input
from rkstreamer.utils.helper import ALBUM_PATTERN
from rkstreamer.types import (
    AlbumModelType,
    AlbumViewType,
    AlbumControllerType,
    CommandType
)


class JioSaavnAlbumController(ControllerUtils):
    """Album Controller implemented for Jio Saavn model"""

    def __init__(self, model: AlbumModelType, view: AlbumViewType) -> None:
        self.model = model
        self.goto_album = model
        self.view = view
        self.commands = {
            ControllerEnum.QUEUE: AlbumQueueCommand(self),
            ControllerEnum.CONTROLS: PlayerControlsCommand(self),
            ControllerEnum.RQUEUE: ReSongQueueCommand(self),
            ControllerEnum.GTALBUM: GotoAlbumCommand(self),
            'as': AlbumSongSelectCommand(self),
            str: AlbumSearchCommand(self),
            int: AlbumSelectCommand(self),
        }
        self.goto_album_songs = {}
        super().__init__(model, view)

    def handle_input(self, user_input: Union[str, int]):
        """Handles the user input"""
        if user_input.startswith('-'):
            re_match = re.match(r'(-\w{1})', user_input)
            try:
                enum_obj = ControllerEnum(re_match.group(1))
                command: CommandType = self.commands.get(enum_obj)
                command.execute(user_input)
            except (ValueError, AttributeError):
                print("Invalid input. Please try again")
        else:
            album_selection = re.match(r'[aA](\d+)', user_input)
            if album_selection:
                song_selection = album_selection.group(1)
                command: CommandType = self.commands.get('as')
                command.execute(song_selection)
            else:
                input_type = type(parse_input(user_input))
                command: CommandType = self.commands.get(input_type)
                command.execute(user_input)


class AlbumSearchCommand(Command):
    """Album Search"""

    def __init__(self, controller: AlbumControllerType):
        self.controller = controller
        self.model: AlbumModelType = self.controller.model
        self.view: AlbumViewType = self.controller.view

    def execute(self, user_input: str):
        match_input = re.match(ALBUM_PATTERN, user_input)
        if match_input:
            album = match_input.group('album').strip()
            match_input.groupdict().pop('album')
            search_results = self.model.search(album, **match_input.groupdict())
            self.view.display(search_results)
        else:
            print("Invalid Input provided!")


class AlbumSelectCommand(Command):
    """Song Select command implementation"""

    def __init__(self, controller: AlbumControllerType):
        self.controller = controller
        self.model: AlbumModelType = self.controller.model
        self.view: AlbumViewType = self.controller.view

    def execute(self, user_input: int):
        album = self.model.select(int(user_input))
        self.view.display_album_songs(album)


class AlbumSongSelectCommand(Command):
    """Song Select command implementation"""

    def __init__(self, controller: AlbumControllerType):
        self.controller = controller
        self.model: AlbumModelType = self.controller.model
        self.view: AlbumViewType = self.controller.view

    def execute(self, user_input: int):
        album_song = self.model.select_song_from_album(int(user_input))
        self.controller.uow_play_songs_remove_loaded(album_song)



"""
Albums provider - API
"""

from html import unescape
from urllib.parse import quote_plus, urlencode
from urllib3 import disable_warnings
from rkstreamer.interfaces.provider import IAlbumProvider
from rkstreamer.utils.helper import LANGUAGES
from rkstreamer.types import (
    AlbumRawType,
    AlbumListRawType,
    NetworkProviderType,
    NetworkProviderResponseType)
disable_warnings()  # Function to suppress the SSL Verification error.


class JioSaavnAlbumProvider(IAlbumProvider):
    """Jio Saavn Album provider API"""

    API_BASE = "https://www.jiosaavn.com/api.php"

    PARAMS_DEFAULT = {'api_version': 4, '_format': 'json', '_marker': 0,
                      'ctx': 'web6dot0'}

    album_search = {'__call': 'search.getAlbumResults',
                    'n': 3, 'p': 1, 'q': ''}

    album_select = {'__call': 'webapi.get', 'token': '',
                    'type': 'album', 'includeMetaTags': 0}

    def __init__(self, client: NetworkProviderType) -> None:
        self.client = client

    def search_albums(self, search_string: str, **kwargs) -> AlbumListRawType:
        self.album_search['q'] = quote_plus(search_string)
        self.album_search['n'] = kwargs.get('num') or 3
        language = [kwargs.get('lang'),] \
            if kwargs.get('lang') \
            else LANGUAGES
        response = self.client.get(
            url=f"{self.API_BASE}?{urlencode(self.album_search|self.PARAMS_DEFAULT, safe='+')}")
        return self._parse_albums(response, lang=language)

    def _parse_albums(self, response: NetworkProviderResponseType, **kwargs) -> AlbumListRawType:
        return [{'name': unescape(album['title']),
                'id': album['perma_url'].split('/')[-1],
                 'music': album['more_info']['music'][:50] if album['more_info']['music'] else '',
                 'artists': unescape(album['subtitle']),
                 'language': album['language'],
                 'song_count': album['more_info']['song_count']}
                for album in response.json()['results'] if album['language'] in kwargs.get('lang')]

    def select_album(self, arg: str, **kwargs) -> AlbumListRawType:
        self.album_select['token'] = arg
        response = self.client.get(
            url=self.API_BASE, params=self.album_select | self.PARAMS_DEFAULT)
        return self._parse_album_songs(response)

    def _parse_album_songs(self, response: NetworkProviderResponseType) -> AlbumListRawType:
        return [{'name': unescape(song['title']),
                 'id': song['id'],
                 'album_name': unescape(song['more_info']['album']),
                 'music': song['more_info']['music'][:50] if song['more_info']['music'] else '',
                 'artists': unescape(song['subtitle'].replace(f" - {song['more_info']['album']}", '')),
                 'duration': song['more_info']['duration'],
                 'token': song['more_info']['encrypted_media_url'],
                 'album_id': song['more_info']['album_url'].split('/')[-1]}
                for song in response.json()['list']]

    def _parse_full_album(self, response: NetworkProviderResponseType) -> AlbumRawType:
        album = response.json()
        return {'name': unescape(album['title']),
                'id': album['perma_url'].split('/')[-1],
                'artists': unescape(album['subtitle']),
                'songs': self._parse_album_songs(response)}

    def select_album_id(self, album_id: str) -> AlbumRawType:
        """Select album using ID"""
        self.album_select['token'] = album_id
        response = self.client.get(
            url=self.API_BASE, params=self.album_select | self.PARAMS_DEFAULT)
        return self._parse_full_album(response)



"""Types used by Views"""

from typing import NewType
from rkstreamer.interfaces import (
    ISongView,
    IAlbumView,
    IPlaylistView
)

SongViewType = NewType('ISongView', ISongView)
AlbumViewType = NewType('IAlbumView', IAlbumView)
PlaylistViewType = NewType('IPlaylistView', IPlaylistView)



"""
Service - Network Provider
"""

from typing import Optional
import requests
from rkstreamer.interfaces.network import INetworkProvider, INetworkProviderResponse


def hook_raise_status(data: INetworkProviderResponse, *args, **kwargs) -> Optional[INetworkProviderResponse]:
    """Check for return status code"""
    try:
        data.raise_for_status()
    except requests.exceptions.HTTPError as exception:
        raise SystemExit(exception) from None
    return data


def requests_wrapper(function) -> INetworkProviderResponse:
    """Wrapper for Request calls"""
    def wrapper(*args, **kwargs):
        try:
            kwargs.update(
                {'hooks': {'response': [hook_raise_status]}})
            response = function(*args, **kwargs)
            return response
        except requests.exceptions.RequestException as exception:
            raise SystemExit(exception) from None
    return wrapper


class PyRequests(INetworkProvider):
    """HTTP Requests bundle for making calls to Stream APIs"""

    def __init__(self, **kwargs) -> None:

        # Creating a Requests Session and passing below parameters to be persistent.
        self.session = requests.session()
        self.session.proxies = kwargs.pop('proxy', None)
        self.session.verify = False
        self.session.headers = {
            'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/106.0.0.0 Safari/537.36'}
        self.kwargs = kwargs
        self.response = {}

    @requests_wrapper
    def get(self, **kwargs):
        self.response = self.session.get(**kwargs | self.kwargs)
        return self.response


class PyResponse(INetworkProviderResponse):
    """HTTP Requests Response"""

    def __init__(self) -> None:
        self.response = requests.Response()

    @property
    def headers(self):
        return self.response.headers

    @property
    def status_code(self):
        return self.response.status_code

    def json(self):
        return self.response.json()

    def raise_for_status(self):
        return self.response.raise_for_status()


__all__ = ['PyRequests']



"""Enum types for Controllers"""

from enum import Enum


class CommandEnum(Enum):
    """Command types"""
    SONG = '-s'
    ALBUM = '-a'
    PLAYLIST = '-p'
    CHART = '-c'


class ControllerEnum(Enum):
    """Song enums"""
    SEARCH = str
    SELECT = int
    QUEUE = '-q'
    CONTROLS = '-c'
    RQUEUE = '-r'
    GTALBUM = '-g'
    PVIEW = '-v'


class QueueEnum(Enum):
    """Song Queue Enum"""
    ADD = 'a'
    REMOVE = 'r'
    PLAY = 'p'


class GotoAlbumEnum(Enum):
    """Goto Album Queue Enum"""
    ADD = 'a'
    PLAY = 'p'



"""Models - Album"""

from typing import Optional
from rkstreamer.interfaces.models import IAlbumModel
from rkstreamer.models.song import JioSaavnSongQueue
from rkstreamer.models.data import Album, AlbumSearch, Song
from rkstreamer.services.album import JioSaavnAlbumProvider
from rkstreamer.services.song import JioSaavnSongProvider
from rkstreamer.types import (
    AlbumType,
    AlbumSearchType,
    AlbumListRawType,
    AlbumSearchIndexType,
    NetworkProviderType,
    SongQueueModelType,
    SongType,
    SongListRawType,
    SongListType,
    SongIndexType
)
from rkstreamer.models.exceptions import (
    InvalidInput
)


class JioSaavnAlbumModel(IAlbumModel):
    """Album model implemented for Jio Saavn service"""

    def __init__(self, network_provider: NetworkProviderType) -> None:
        self.network_provider = network_provider
        self.stream_provider = JioSaavnAlbumProvider(
            client=self.network_provider)
        self.song_provider = JioSaavnSongProvider(client=self.network_provider)
        self.queue: SongQueueModelType = JioSaavnSongQueue()
        self.indexed_search_albums = {}
        self.indexed_album_songs = {}
        self._recomm_song_index = 1

    def _create_song(self, **kwargs) -> SongType:
        return Song(**kwargs)

    def _create_search_album(self, **kwargs) -> AlbumSearchType:
        return AlbumSearch(**kwargs)

    def _create_album(self, **kwargs) -> AlbumType:
        return Album(**kwargs)

    def _create_search_album_index(self, albums: AlbumListRawType) -> AlbumSearchIndexType:
        self.indexed_search_albums = {count: self._create_search_album(**album)
                                      for count, album in enumerate(albums, 1)}
        return self.indexed_search_albums

    def _create_album_song_index(self, albums: AlbumListRawType) -> SongIndexType:
        self.indexed_album_songs = {count: self._create_song(**album)
                                    for count, album in enumerate(albums, 1)}
        return self.indexed_album_songs

    def _create_recomm_song(self, songs: SongListRawType) -> SongIndexType:
        recomm_songs = {}
        _songs: SongListType = [self._create_song(**song) for song in songs]
        for song in _songs:
            if song not in self.queue.rsongs_copy_:
                recomm_songs.update({self._recomm_song_index: song})
                self._recomm_song_index += 1
        return recomm_songs

    def search(self, search_string: str, **kwargs) -> AlbumSearchIndexType:
        search_result = self.stream_provider.search_albums(
            search_string, **kwargs)
        return self._create_search_album_index(search_result)

    def select(self, selection: int, **kwargs) -> AlbumType:
        selected_album = self.indexed_search_albums.get(int(selection))
        if selected_album:
            album_songs_raw = self.stream_provider.select_album(
                selected_album.id, **kwargs)
            selected_album.__dict__.update(
                {'songs': self._create_album_song_index(album_songs_raw)})
            return self._create_album(**selected_album.__dict__)
        raise InvalidInput("Invalid album selection input provided.")

    def select_album_using_id(self, album_id: str) -> AlbumType:
        """Select album using ID"""
        album_songs_raw = self.stream_provider.select_album_id(album_id)
        album_songs_raw.update(
            {'songs': self._create_album_song_index(album_songs_raw['songs'])})
        return self._create_album(**album_songs_raw)

    def select_song_from_album(self, selection: int) -> Optional[SongType]:
        """Selecting song from the album"""
        selected_song = self.indexed_album_songs.get(selection)
        if selected_song:
            song_url = self.song_provider.select_song(selected_song.token)
            if song_url:
                selected_song.__dict__.update({'stream_url': song_url})
                selected_song.__dict__.update({'status': 'Loaded'})
                return selected_song
        raise InvalidInput("Invalid album song selection input provided.")

    def get_related_songs(self, data: str) -> SongIndexType:
        """Gets recommended songs using song_id and updates the RQueue"""
        recomm_songs_raw: SongListRawType = self.song_provider.get_recomm_songs(
            data)
        if recomm_songs_raw:
            return self._create_recomm_song(recomm_songs_raw)

    def get_song_url(self, data: str) -> str:
        """Get the song's stream url using Enc Url Token - used for rsongs download"""
        stream_url = self.song_provider.select_song(data)
        return stream_url



"""
Design Patterns for Interfaces
"""

from ._interface import ABC, abstractmethod


class Command(ABC):
    """Command pattern interface"""

    @abstractmethod
    def execute(self, user_input):
        """Execute method that does the action"""



"""
Types used by domain models
"""

from typing import NewType
from rkstreamer.models.data import (
    Song,
    SongSearch,
    SongQueue,
    Album,
    AlbumSearch,
    Playlist,
    PlaylistSearch)
from rkstreamer.interfaces import (
    ISongModel,
    ISongQueue,
    IAlbumModel,
    IPlaylistModel,
    ISongProvider,
    IAlbumProvider,
    IPlaylistProvider)

# Interfaces

SongModelType = NewType('ISongModel', ISongModel)
SongQueueModelType = NewType('ISongQueue', ISongQueue)
AlbumModelType = NewType('IAlbumModel', IAlbumModel)
PlaylistModelType = NewType('IPlaylistModel', IPlaylistModel)

# Primitives

SongRawType = NewType('SongRaw', dict[str,str])
SongListRawType = NewType('SongListRaw', list[SongRawType])
AlbumRawType = NewType('AlbumRaw', dict[str,str])
AlbumListRawType = NewType('SongListRaw', list[AlbumRawType])
PlaylistRawType = NewType("PlaylistRaw", dict[str,str])
PListRawType = NewType("PlaylistListRaw", list[PlaylistRawType])

# Data

SongType = NewType('Song', Song)
SongSearchType = NewType('SongSearch', SongSearch)
SongQueueType = NewType('SongQueue', SongQueue)
SongListType = NewType('SongList', list[SongType])
SongIndexType = NewType('SongIndex', dict[int, SongType])
SongUrl = NewType('SongUrl', str)
SongSearchIndexType = NewType('SongSearchIndex', dict[int,SongSearchType])
SongQueueIndexType = NewType('SongQueueIndex', dict[int,SongQueueType])

AlbumType = NewType('Album', Album)
AlbumSearchType = NewType('AlbumSearch', AlbumSearch)
AlbumSearchIndexType = NewType('AlbumSearchIndex', dict[int,AlbumSearchType])

PlaylistType = NewType('Playlist', Playlist)
PlaylistSearchType = NewType('PlaylistSearch', PlaylistSearch)
PlaylistSearchIndexType = NewType('AlbumSearchIndex', dict[int,PlaylistSearchType])

SongProviderType = NewType("SongProvider", ISongProvider)
AlbumProviderType = NewType("AlbumProvider", IAlbumProvider)
PlaylistProviderType = NewType("PlaylistProvider", IPlaylistProvider)



"""
Song view Implementation module
"""

from typing import Callable
from rkstreamer.interfaces.views import ISongView
from rkstreamer.types import (
    SongSearchIndexType,
    SongType,
    SongQueueIndexType,
    MusicPlayerControlsType,
    )


class JioSaavnSongView(ISongView):
    """Jio Saavn Songs View"""

    def __init__(self, player: MusicPlayerControlsType) -> None:
        self.player = player
        self.media_player = self.player.mlplayer_factory

    def display(self, entity: SongSearchIndexType) -> None:
        print('\n\033[33m\033[01m***Songs Search***\033[0m\n')
        for count, song in entity.items():
            print(f"{count}- {song.name} ({song.language})")
            print(f"Album - {song.album_name}")
            print(f"Artists - {song.artists}")
            print()

    def display_selected(self, entity: SongType) -> None:
        print(entity)

    def display_queue(self, queue: SongQueueIndexType) -> None:
        print('\n\033[33m\033[01m***Songs Queue***\033[0m\n')
        for count, song in queue.items():
            print(f"#{count} - {song.name} - {song.status}")
        print()

    def display_rsongs_queue(self, queue: list[SongType]) -> None:
        """Displays the recommended songs queue"""
        print('\n\033[33m\033[01m***Recommended Songs***\033[0m\n')
        for count, song in queue.items():
            print(f"#{count} - {song.name} - {song.artists}")
        print()

    def display_gotoalbum_songs(self, entity) -> None:
        """Display album songs"""
        print('\n\033[33m\033[01m***Albums Songs***\033[0m\n')
        print(f'\033[04m\033[01m\033[32mALBUM: {entity.name}\033[0m\n')
        for count, song in entity.songs.items():
            print(f"{count}- {song.name}")
        print()

    def play_media(self, media: SongType) -> None:
        # print(f"\n\033[31m> Playing '{media.name}' <\033[0m\n ")
        return self.media_player.play_media(media.stream_url)

    def remove_media(self, media: SongType) -> None:
        """Removes media from media list"""
        return self.media_player.remove_media(media.stream_url)

    def add_media(self, media: SongType) -> None:
        """Adding media to media list"""
        return self.media_player.add_media(media.stream_url)

    def player_input(self, user_input: str) -> None:
        """Player input"""
        self.player.player_controls(user_input)

    def set_controller_callback(self, callback_fn: Callable) -> None:
        """Updates the callback attribute to mlplayer's monitor state"""
        self.player.monitor_state.callback = callback_fn

    def play(self):
        """Play the media list"""
        return self.player.mlplayer_controls.play()

    def stop(self):
        """stop the media list"""
        return self.player.mlplayer_controls.stop()



"""
Playlist view implementation module
"""

from rkstreamer.types import MusicPlayerControlsType
from rkstreamer.views.song import JioSaavnSongView


class JioSaavnPlaylistView(JioSaavnSongView):
    """Jio Saavn Playlist view"""

    def __init__(self, player: MusicPlayerControlsType) -> None:
        self.player = player
        super().__init__(player)

    def display(self, entity) -> None:
        print('\n\033[33m\033[01m***Playlist Search***\033[0m\n')
        for count, playlist in entity.items():
            print(f"{count}- {playlist.name} ({playlist.language})")
            print(f"Song Count - {playlist.song_count}")
            print()

    def display_playlist_songs(self, entity) -> None:
        """Display playlist songs"""
        print('\n\033[33m\033[01m***Playlist Songs***\033[0m\n')
        for count, song in enumerate(entity.songs):
            print(f"#{count}- {song.name}")
        print()

    def add_media_list(self, songs_list: list):
        """Add songs list to player mlist"""
        return self.player.mlplayer_factory.add_medias(songs_list)



"""Controller Implementation module"""

import re
from typing import Union
from rkstreamer.interfaces.patterns import Command
from rkstreamer.controllers.enums import ControllerEnum
from rkstreamer.controllers.queue import SongQueueCommand, ReSongQueueCommand
from rkstreamer.controllers.patterns import (
    ControllerUtils,
    PlayerControlsCommand,
    GotoAlbumCommand
)
from rkstreamer.models.album import JioSaavnAlbumModel
from rkstreamer.utils.helper import parse_input
from rkstreamer.utils.helper import SONG_PATTERN
from rkstreamer.types import (
    SongModelType,
    SongViewType,
    SongControllerType,
    CommandType
)


class JioSaavnSongController(ControllerUtils):
    """Song Controller implemented for Jio Saavn model"""

    def __init__(self, model: SongModelType, view: SongViewType) -> None:
        self.model = model
        self.view = view
        self.goto_album = JioSaavnAlbumModel(
            network_provider=model.network_provider)
        self.commands = {
            ControllerEnum.QUEUE: SongQueueCommand(self),
            ControllerEnum.CONTROLS: PlayerControlsCommand(self),
            ControllerEnum.RQUEUE: ReSongQueueCommand(self),
            ControllerEnum.GTALBUM: GotoAlbumCommand(self),
            str: SongSearchCommand(self),
            int: SongSelectCommand(self),
        }
        self.goto_album_songs = {}
        super().__init__(model, view)

    def handle_input(self, user_input: Union[str, int]):
        if user_input.startswith('-'):
            re_match = re.match(r'(-\w{1})', user_input)
            try:
                enum_obj = ControllerEnum(re_match.group(1))
                command: CommandType = self.commands.get(enum_obj)
                command.execute(user_input)
            except (ValueError, AttributeError) as exc:
                print(f"{exc.__class__, 'Invalid input. Please try again'}")
        else:
            input_type = type(parse_input(user_input))
            command: CommandType = self.commands.get(input_type)
            command.execute(user_input)


class SongSearchCommand(Command):
    """Song Search command implementation"""

    def __init__(self, controller: SongControllerType):
        self.controller = controller
        self.model: SongModelType = self.controller.model
        self.view: SongViewType = self.controller.view

    def execute(self, user_input: str):
        match_input = re.match(SONG_PATTERN, user_input)
        if match_input:
            song = match_input.group('song').strip()
            match_input.groupdict().pop('song')
            search_results = self.model.search(song, **match_input.groupdict())
            self.view.display(search_results)
        else:
            print("Invalid Input provided!")


class SongSelectCommand(Command):
    """Song Select command implementation"""

    def __init__(self, controller: SongControllerType):
        self.controller = controller
        self.model: SongModelType = self.controller.model
        self.view: SongViewType = self.controller.view

    def execute(self, user_input: int):
        song = self.model.select(user_input)
        self.controller.uow_play_songs_remove_loaded(song)



"""Type used by Controllers"""

from typing import NewType
from rkstreamer.interfaces import (
    ISongController,
    IAlbumController,
    IPlaylistController
)

SongControllerType = NewType('ISongController', ISongController)
AlbumControllerType = NewType('IAlbumController', IAlbumController)
PlaylistControllerType = NewType('IPlaylistController', IPlaylistController)



"""Test - Play song"""

from rkstreamer.controllers.song import JioSaavnSongController
from rkstreamer.models.song import JioSaavnSongModel
from rkstreamer.views.song import JioSaavnSongView
from rkstreamer.services.network import PyRequests

pyrequests = PyRequests()
test = JioSaavnSongController(model=JioSaavnSongModel(pyrequests), view=JioSaavnSongView())

while True:
    user_input = input("Enter the song name: ")
    test.handle_input(user_input)


"""
Design patterns implementation
"""

from __future__ import annotations
import threading
from typing import Union, TYPE_CHECKING
from rkstreamer.interfaces.controllers import ISongController
from rkstreamer.interfaces.patterns import Command
from rkstreamer.controllers.enums import GotoAlbumEnum
if TYPE_CHECKING:
    from rkstreamer.types import (
        SongControllerType,
        SongType,
        SongModelType,
        SongViewType,
        AlbumModelType,
        AlbumViewType,
        CommandType
    )


class ControllerUtils(ISongController):
    """Generic controller utils for handling songs"""

    def __init__(
            self,
            model: Union[SongModelType, AlbumModelType],
            view: Union[SongViewType, AlbumViewType]) -> None:

        self.model = model
        self.view = view

        self.view.set_controller_callback(self.uow_update_song_status)

        _ = threading.Timer(30, self.monitor_queue_pull_rsong)
        _.setDaemon(True)
        _.start()

    def uow_update_song_status(self, status: str, stream_url: str):
        """Updating song status in queue to "Played"
        if the song has been selected to play from search or from queue
        This function fetchs the rsongs for the playing song and updates rsong list."""
        song = self.model.queue.update_qstatus(status, stream_url)
        if song and len(self.model.queue.get_rsongs) <= 50:
            self.uow_add_rsongs_rqueue(song.id)

    def monitor_queue_pull_rsong(self):
        """Pull rsong by monitoring the queue songs status"""
        if self.model.queue.check_status(self.model.queue.get_queue):
            get_rsong = self.model.queue.pop_rsong()
            if get_rsong:
                get_rsong.stream_url = self.model.get_song_url(get_rsong.token)
                self.uow_add_songs_queue(get_rsong)
        threading.Timer(45, self.monitor_queue_pull_rsong).start()

    def uow_add_songs_queue(self, song: SongType):
        """UOW: Add songs to queue & media list. Doesn't play it"""
        self.model.queue.add(song)
        self.view.add_media(song)

    def uow_add_rsongs_rqueue(self, data: str):
        """UOW: Add Recommended songs to RQueue
        :data - song_id"""
        recomm_songs = self.model.get_related_songs(data)
        if recomm_songs:
            self.model.queue.update_rqueue(recomm_songs)

    def uow_play_songs_remove_loaded(self, song: SongType):
        """UOW: changes the queue status to played &
        play the media - suitable for rsongs play"""
        self.model.queue.change_loaded_status()
        self.model.queue.add(song)
        self.view.play_media(song)

    def uow_play_songs_remove_loaded_before(self, song: SongType):
        """UOW: changes the status of queue song to played (only those are before this song)"""
        self.model.queue.change_loaded_status_before(song)
        self.view.play_media(song)


class GotoAlbumCommand(Command):
    """Goto Album Command"""

    def __init__(self, controller: SongControllerType):
        self.controller = controller
        self._model = controller.model
        self.model: AlbumModelType = self.controller.goto_album
        self.view: SongViewType = self.controller.view
        self.commands = {
            GotoAlbumEnum.ADD: GotoAlbumAddCommand(self.controller),
            GotoAlbumEnum.PLAY: GotoAlbumPlayCommand(self.controller),
        }
        self.current_album_id = ''

    def execute(self, user_input):
        if user_input == '-g':
            album_id = self._model.queue.current_playing_song.album_id
            if self.current_album_id != album_id:
                # self.controller.goto_album_songs.clear() # flush dict before saving new songs.
                self.controller.goto_album_songs = self.model.select_album_using_id(album_id)
                self.current_album_id = album_id
                self.view.display_gotoalbum_songs(self.controller.goto_album_songs)
            else:
                self.view.display_gotoalbum_songs(self.controller.goto_album_songs)
        elif len(user_input) > 2:
            user_input = user_input.replace('-g', '')
            try:
                enum_obj = GotoAlbumEnum(user_input[0])
                command: CommandType = self.commands.get(enum_obj)
                command.execute(user_input.replace(
                    user_input[0], '').split(','))
            except (ValueError, AttributeError) as exc:
                print(f"{exc.__class__, 'Invalid input. Please try again'}")


class GotoAlbumAddCommand(Command):
    """Goto Album - Add Command"""

    def __init__(self, controller: SongControllerType):
        self.controller = controller
        self.model: SongModelType = self.controller.model
        self.view: SongViewType = self.controller.view

    def execute(self, user_input: list):
        for number in user_input:
            album_song = self.controller.goto_album_songs.songs.get(
                int(number))
            if album_song:
                album_song.stream_url = self.model.get_song_url(album_song.token)
                album_song.status = 'Loaded'
                self.controller.uow_add_songs_queue(album_song)


class GotoAlbumPlayCommand(Command):
    """Goto Album - Play Command"""

    def __init__(self, controller: SongControllerType):
        self.controller = controller
        self.model: SongModelType = self.controller.model
        self.view: SongViewType = self.controller.view

    def execute(self, user_input: list):
        for number in user_input:
            album_song = self.controller.goto_album_songs.songs.get(
                int(number))
            if album_song:
                album_song.stream_url = self.model.get_song_url(album_song.token)
                album_song.status = 'Loaded'
                self.controller.uow_play_songs_remove_loaded(album_song)


class PlayerControlsCommand(Command):
    """Player Controls"""

    def __init__(self, controller: SongControllerType):
        self.controller = controller
        self.model: SongModelType = self.controller.model
        self.view: SongViewType = self.controller.view

    def execute(self, user_input: str):
        if user_input.lower().startswith('-c'):
            self.view.player_input(user_input.replace('-c', ''))



"""
Dependency Injection Modules
"""

from .models import *
from .controllers import *
from .views import *
from .network import *
from .patterns import *
from .player import *
from .provider import *



"""
Songs Provider API
"""

import random
from html import unescape
from urllib.parse import quote_plus, urlencode
from urllib3 import disable_warnings
from rkstreamer.interfaces.provider import ISongProvider
from rkstreamer.utils.helper import LANGUAGES
from rkstreamer.types import (
    SongListRawType,
    NetworkProviderType,
    NetworkProviderResponseType
)
disable_warnings()  # Function to suppress the SSL Verification error.


class JioSaavnSongProvider(ISongProvider):
    """Jio Saavn - Song provider API"""

    API_BASE = "https://www.jiosaavn.com/api.php"

    PARAMS_DEFAULT = {'api_version': 4, '_format': 'json', '_marker': 0,
                      'ctx': 'web6dot0'}

    # n - num of results, q- search string
    song_search = {'__call': 'search.getResults', 'p': 1, 'q': '', 'n': ''}

    song_download = {'__call': 'song.generateAuthToken',
                     'url': '', 'bitrate': ''}  # url - enc media url, bitrate - 160 or 320

    entity_station = {'__call': 'webradio.createEntityStation', 'entity_id': '',
                      'entity_type': 'queue', 'freemium': '', 'shared': ''}

    # k - count of rsongs.
    recomm_songs = {'__call': 'webradio.getSong', 'stationid': '', 'k': ''}

    def __init__(self, client: NetworkProviderType) -> None:
        self.client = client

    def search_songs(self, search_string: str, **kwargs) -> SongListRawType:
        """Search songs using the search string"""
        self.song_search['q'] = quote_plus(search_string)
        self.song_search['n'] = kwargs.get('num') or 3
        self.song_download['bitrate'] = kwargs.get('bitrate') or 320
        language = [kwargs.get('lang'),] \
            if kwargs.get('lang') \
            else LANGUAGES
        response = self.client.get(
            url=f"{self.API_BASE}?{urlencode(self.song_search|self.PARAMS_DEFAULT, safe='+')}")
        return self._parse_songs(response, lang=language)

    def _parse_songs(self, response: NetworkProviderResponseType, **kwargs) -> SongListRawType:
        """Parsing songs info from search songs call"""
        return [{'name': unescape(song['title']),
                'id': song['id'],
                 'album_name': unescape(song['more_info']['album']),
                 'music': song['more_info']['music'][:50] if song['more_info']['music'] else '',
                 'artists': unescape(song['subtitle'].replace(f" - {song['more_info']['album']}", '')),
                 'language': song['language'],
                 'duration': song['more_info']['duration'],
                 'token': song['more_info']['encrypted_media_url'],
                 'album_id': song['more_info']['album_url'].split('/')[-1]}
                for song in response.json()['results'] if song['language'] in kwargs.get('lang')]

    def select_song(self, arg: str, **kwargs) -> str:
        self.song_download['url'] = arg  # Encrypted URL
        response = self.client.get(
            url=self.API_BASE, params=self.song_download | self.PARAMS_DEFAULT)
        return self._parse_song_url(response)

    def _parse_song_url(self, response: NetworkProviderResponseType) -> str:
        """Get the song download URL"""
        auth_url = response.json()['auth_url']
        stream_url = self.client.get(url=auth_url, allow_redirects=False)
        return stream_url.headers['Location']

    def _get_station_id(self, song_id: str) -> str:
        self.entity_station['entity_id'] = f'["{song_id}"]'
        sid_response = self.client.get(
            url=self.API_BASE, params=self.entity_station | self.PARAMS_DEFAULT)
        sid = sid_response.json()['stationid']
        return sid

    def _parse_recomm_songs(self, response: NetworkProviderResponseType) -> SongListRawType:
        try:
            return [{'name': unescape(song['song']['title']),
                    'id': song['song']['id'],
                     'album_name': unescape(song['song']['more_info']['album']),
                     'music': song['song']['more_info']['music'][:50] if song ['song']['more_info']['music'] else '',
                     'artists': unescape(song['song']['subtitle'].replace(f" - {song['song']['more_info']['album']}", '')),
                     'duration': song['song']['more_info']['duration'],
                     'token': song['song']['more_info']['encrypted_media_url'],
                     'album_id': song['song']['more_info']['album_url'].split('/')[-1]}
                    for key, song in response.json().items() if key != 'stationid']
        except (KeyError, IndexError, TypeError):
            print("Get Rsongs failed")

    def get_recomm_songs(self, song_id: str, **kwargs) -> SongListRawType:
        """Get recommended songs from song_id"""
        station_id = self._get_station_id(song_id)
        self.recomm_songs['stationid'] = station_id
        self.recomm_songs['k'] = kwargs.get('rsongs') or random.randint(10, 15)
        recomm_songs = self.client.get(
            url=self.API_BASE, params=self.recomm_songs | self.PARAMS_DEFAULT)
        return self._parse_recomm_songs(recomm_songs)



"""Tests: Get album using ID"""

from rkstreamer.services.network import PyRequests
from rkstreamer.models.album import JioSaavnAlbumModel


test = JioSaavnAlbumModel(network_provider=PyRequests(verify=False))
search = test.select_album_using_id('qKErkhPpdTE_')
print(search)


"""Tests: Search playlist"""

from rkstreamer.services.network import PyRequests
from rkstreamer.models.playlist import JioSaavnPlaylistModel


proxy = {'https': 'http://127.0.0.1:8888'}
test = JioSaavnPlaylistModel(network_provider=PyRequests(verify=False, proxy=proxy))
search = test.search('akon')
print(search)


"""Tests: Playlists"""

from rkstreamer.services.player import PyVLCPlayer
from rkstreamer.controllers.playlist import JioSaavnPlaylistController
from rkstreamer.models.playlist import JioSaavnPlaylistModel
from rkstreamer.views.playlist import JioSaavnPlaylistView
from rkstreamer.services.network import PyRequests

proxy = {'https': 'http://127.0.0.1:8888'}
pyrequests = PyRequests(proxy=None)

test = JioSaavnPlaylistController(
    model=JioSaavnPlaylistModel(pyrequests),
    view=JioSaavnPlaylistView(player=PyVLCPlayer()))

while True:
    user_input = input("Enter the playlist name: ")
    if user_input:
        test.handle_input(user_input)



"""
DI module for Views
"""

from ._interface import ABC, abstractmethod


class IView(ABC):
    """Interface for Song View"""

    @abstractmethod
    def display(self, entity):
        """Displays the entity"""

    @abstractmethod
    def display_selected(self, entity):
        """Displays the entity"""

    @abstractmethod
    def play(self):
        """Play the media list"""

    @abstractmethod
    def stop(self):
        """stop the media list"""


class ISongView(IView):
    """Interface for Song View"""

    @abstractmethod
    def display_queue(self, queue):
        """Displays the queue"""
    
    @abstractmethod
    def display_rsongs_queue(self, queue):
        """Displays the related songs queue"""

    @abstractmethod
    def play_media(self, media) -> None:
        """Play the given media"""

    @abstractmethod
    def remove_media(self, media) -> None:
        """Play the given media"""

    @abstractmethod
    def add_media(self, media) -> None:
        """Play the given media"""

    @abstractmethod
    def player_input(self, user_input: str):
        """Input to music player"""

    @abstractmethod
    def set_controller_callback(self, callback_fn):
        """Updates the callback attribute to mlplayer's monitor state"""


class IAlbumView(ISongView):
    """Album View"""

    @abstractmethod
    def display_album_songs(self, entity):
        """Display album songs"""

class IPlaylistView(ISongView):
    """Playlist view"""

    @abstractmethod
    def display_playlist_songs(self, entity):
        """Display playlist songs"""

    @abstractmethod
    def add_media_list(self, songs_list):
        """Add songs list to player media list"""




"""Queue command module"""

from __future__ import annotations
from typing import TYPE_CHECKING
from rkstreamer.interfaces.patterns import Command
from rkstreamer.controllers.enums import QueueEnum
if TYPE_CHECKING:
    from rkstreamer.types import (
        SongControllerType,
        SongModelType,
        SongViewType,
        AlbumControllerType,
        AlbumModelType,
        AlbumViewType,
        CommandType
    )

class SongQueueCommand(Command):
    """Song Queue command implementation"""

    def __init__(self, controller: SongControllerType):
        self.controller = controller
        self.model: SongModelType = self.controller.model
        self.view: SongViewType = self.controller.view
        self.commands = {
            QueueEnum.ADD: SongQueueAddCommand(self.controller),
            QueueEnum.REMOVE: SongQueueRemoveCommand(self.controller),
            QueueEnum.PLAY: SongQueuePlayCommand(self.controller),
        }

    def execute(self, user_input: str):
        if user_input == '-q':
            queue = self.model.queue.get_indexed_queue
            self.view.display_queue(queue)
        elif len(user_input) > 2:
            user_input = user_input.replace('-q', '')
            try:
                enum_obj = QueueEnum(user_input[0])
                command: CommandType = self.commands.get(enum_obj)
                command.execute(user_input.replace(
                    user_input[0], '').split(','))
            except (ValueError, AttributeError) as exc:
                print(f"{exc.__class__, 'Invalid input. Please try again'}")


class SongQueueAddCommand(Command):
    """Song Queue - Add command"""

    def __init__(self, controller: SongControllerType):
        self.controller = controller
        self.model: SongModelType = self.controller.model

    def execute(self, user_input: list):
        for number in user_input:
            song = self.model.select(int(number))
            self.controller.uow_add_songs_queue(song)


class SongQueueRemoveCommand(Command):
    """Song Queue - Remove command"""

    def __init__(self, controller: SongControllerType):
        self.controller = controller
        self.model: SongModelType = self.controller.model
        self.view: SongViewType = self.controller.view

    def execute(self, user_input: list):
        songs = self.model.queue.remove(user_input)
        for song in songs:
            self.view.remove_media(song)


class SongQueuePlayCommand(Command):
    """Song Queue - Play command"""

    def __init__(self, controller: SongControllerType):
        self.controller = controller
        self.model: SongModelType = self.controller.model
        self.view: SongViewType = self.controller.view

    def execute(self, user_input: list):
        song = self.model.queue.fetch(user_input[0])
        self.controller.uow_play_songs_remove_loaded_before(song)


class AlbumQueueCommand(Command):
    """Album queue command"""

    def __init__(self, controller: AlbumControllerType):
        self.controller = controller
        self.model: AlbumModelType = self.controller.model
        self.view: AlbumViewType = self.controller.view
        self.commands = {
            QueueEnum.ADD: AlbumQueueAddCommand(self.controller),
            QueueEnum.REMOVE: SongQueueRemoveCommand(self.controller),
            QueueEnum.PLAY: SongQueuePlayCommand(self.controller)
        }

    def execute(self, user_input: str):
        if user_input == '-q':
            queue = self.model.queue.get_indexed_queue
            self.view.display_queue(queue)
        elif len(user_input) > 2:
            user_input = user_input.replace('-q', '')
            try:
                enum_obj = QueueEnum(user_input[0])
                command: CommandType = self.commands.get(enum_obj)
                command.execute(user_input.replace(
                    user_input[0], '').split(','))
            except (ValueError, AttributeError) as exc:
                print(f"{exc.__class__, 'Invalid input. Please try again'}")


class AlbumQueueAddCommand(Command):
    """Album Queue - Add command"""

    def __init__(self, controller: SongControllerType):
        self.controller = controller
        self.model: AlbumModelType = self.controller.model

    def execute(self, user_input: list):
        for number in user_input:
            song = self.model.select_song_from_album(int(number))
            self.controller.uow_add_songs_queue(song)


class ReSongQueueCommand(Command):
    """Recommended songs - Queue Command"""

    def __init__(self, controller: SongControllerType):
        self.controller = controller
        self.model: SongModelType = self.controller.model
        self.view: SongViewType = self.controller.view
        self.commands = {
            QueueEnum.ADD: ReSongQueueAddCommand(self.controller),
            QueueEnum.REMOVE: ReSongQueueRemoveCommand(self.controller),
            QueueEnum.PLAY: ReSongQueuePlayCommand(self.controller)
        }

    def execute(self, user_input: str):
        if user_input == '-r':
            queue = self.model.queue.get_rsongs
            self.view.display_rsongs_queue(queue)
        elif len(user_input) > 2:
            user_input = user_input.replace('-r', '')
            try:
                enum_obj = QueueEnum(user_input[0])
                command: CommandType = self.commands.get(enum_obj)
                command.execute(user_input.replace(
                    user_input[0], '').split(','))
            except (ValueError, AttributeError) as exc:
                print(f"{exc.__class__, 'Invalid input. Please try again'}")


class ReSongQueueAddCommand(Command):
    """RS Queue - Add Command"""

    def __init__(self, controller: SongControllerType):
        self.controller = controller
        self.model: SongModelType = self.controller.model
        self.view: SongViewType = self.controller.view

    def execute(self, user_input: list):
        for number in user_input:
            rsong = self.model.queue.get_rsong_index(int(number))
            rsong.stream_url = self.model.get_song_url(rsong.token)
            rsong.status = 'Loaded'
            self.controller.uow_add_songs_queue(rsong)


class ReSongQueueRemoveCommand(Command):
    """RS Queue - Remove Command"""

    def __init__(self, controller: SongControllerType):
        self.controller = controller
        self.model: SongModelType = self.controller.model
        self.view: SongViewType = self.controller.view

    def execute(self, user_input: list):
        for index in user_input:
            self.model.queue.remove_rsong_index(int(index))


class ReSongQueuePlayCommand(Command):
    """RS Queue - Play command"""

    def __init__(self, controller: SongControllerType):
        self.controller = controller
        self.model: SongModelType = self.controller.model
        self.view: SongViewType = self.controller.view

    def execute(self, user_input: list):
        song = self.model.queue.get_rsong_index(int(user_input[0]))
        if song:
            song.stream_url = self.model.get_song_url(song.token)
            song.status = 'Loaded'
            self.controller.uow_play_songs_remove_loaded(song)



"""
Services - Music Player Provider
"""

import threading
from typing import Optional, Callable
from vlc import Instance, MediaListPlayer, MediaList, State, MediaPlayer, Media
from rkstreamer.interfaces.player import MusicPlayer, MusicPlayerControls


class PyVLCPlayerInstance(MusicPlayer):
    """VLC Player implementation"""

    def __init__(self) -> None:
        self.songs_list = []
        self.instance: Instance = Instance()
        self.player: MediaListPlayer = self.instance.media_list_player_new()
        self.media_list: MediaList = None
        self._init_media_list()

    def _init_media_list(self):
        """Creates a new media list adds to MLplayer"""
        self.media_list = self.instance.media_list_new()
        self.player.set_media_list(self.media_list)

    def _set_new_media(self):
        """Sets the songs_list as new media list"""
        self.media_list = self.instance.media_list_new(self.songs_list)
        self.player.set_media_list(self.media_list)

    def add_media(self, media_url: str):
        """Appends the new media to media_list"""
        if media_url not in self.songs_list:
            self.songs_list.append(media_url)
            self._set_new_media()

    def add_medias(self, media_urls: list):
        """Appends the new media from LIST to media_list."""
        self.songs_list = media_urls #this is required for play_media.
        self.media_list = self.instance.media_list_new(self.songs_list)
        self.player.set_media_list(self.media_list)

    def remove_media(self, media_url: str):
        """Removes media from media list and sets the updated list"""
        if media_url in self.songs_list:
            self.songs_list.remove(media_url)
            self._set_new_media()

    def play_media(self, media_url: str):
        """Plays the media from the song list.
        It checks if the media_url is already part of song_list.
        If yes, then play that media by getting the index.
        If not, then append it to the media_list, set it and then play the added song."""
        if media_url in self.songs_list:
            song_index = self.songs_list.index(media_url)
            return self.player.play_item_at_index(song_index)
        else:
            self.add_media(media_url)
            self._set_new_media()
            return self.player.play_item_at_index(self.songs_list.index(media_url))

    @property
    def get_media_player(self):
        """Gets the instance of Mplayer used"""
        return self.player.get_media_player()


class PyVLCPlayer(MusicPlayerControls):
    """Handler for Player Controls"""

    def __init__(
            self,
            mlplayer_factory: Optional[PyVLCPlayerInstance] = None) -> None:

        self.mlplayer_factory = mlplayer_factory if mlplayer_factory else PyVLCPlayerInstance()
        self.mlplayer_controls = MediaListPlayerControls(self.mlplayer_factory.player)
        self.mplayer_controls = MediaPlayerControls(self.mlplayer_factory.get_media_player)
        # Media List player will invoke MediaPlayerControls and pass it other controls.
        self.volume_controls = VolumeControls(self.mplayer_controls)
        self.state_controls = StateControls(self.mplayer_controls)
        self.misc_controls = MiscControls(self.mplayer_controls)
        self.playback_controls = PlaybackControls(self.mplayer_controls)
        self.monitor_state = MonitorState(self.mplayer_controls)

    def player_controls(self, user_input: str):
        """Handler method for Player Controls"""

        if user_input.lower().startswith(('+', '-', 'v')):
            self.volume_controls.manage(user_input)

        elif user_input.lower().startswith(('n','pr')):
            if user_input == 'n':
                self.mlplayer_controls.next_song()
            elif user_input == 'pr':
                self.mlplayer_controls.previous_song()

        elif user_input.lower().startswith(('p', 'q', 'e', 's')):
            self.state_controls.manage(user_input)

        elif user_input.lower().startswith(('lo', 'co', 't')):
            self.misc_controls.manage(user_input)

        elif user_input.lower().startswith(('<', '>')):
            self.playback_controls.manage(user_input)


class MediaPlayerControls():
    """Player Controls"""

    def __init__(self, media_player_instance: MediaPlayer) -> None:
        self.media = media_player_instance
        self.devices = []
        self.volume = 100
        self.state = None
        self.get_devices()

    @property
    def get_song_url_from_player(self):
        """Get Media URL from MediaPlayer instance"""
        media_playing: Media = self.media.get_media()
        return media_playing.get_mrl()

    def get_devices(self) -> list:
        """Gets the available audio output devices"""
        mods = self.media.audio_output_device_enum()
        count = 0
        if mods:
            mod = mods
            while mod:
                mod = mod.contents
                self.devices.append(
                    {'id': count, 'device': mod.device,
                        'description': (mod.description).decode()}
                )
                count += 1
                mod = mod.next

    def display_devices(self) -> None:
        """Pretty print list of output devices."""
        for device in self.devices:
            print(f"ID: {device['id']}")
            print(f"Device: {device['description']}")
        print()

    def select_device(self, selection_: int) -> None:
        """Selects the output device."""
        for device in self.devices:
            if device['id'] == selection_:
                print(f'Switching output to "{device["description"]}"')
                self.media.audio_output_device_set(None, device['device'])
                break
        print()

    def get_volume(self) -> int:
        """Get current volume value"""
        setattr(self, "volume", self.media.audio_get_volume())
        return self.volume

    def set_volume(self, volume: int) -> int:
        """Set volume between 0 - 100"""
        if volume < 0:
            print("Cannot set volume below 0")
        elif volume <= 100:
            self.media.audio_set_volume(volume)
            self.volume = volume
        else:
            print("Cannot set volume above 100")

    def increase_volume(self, volume: int) -> None:
        """Increase volume"""
        self.get_volume()
        if self.volume < 100:
            if (self.volume+volume) <= 100:
                self.media.audio_set_volume(
                    self.volume+volume
                )
                print(f"Volume: {self.media.audio_get_volume()}")
            else:
                self.media.audio_set_volume(100)
                self.volume = 100

    def decrease_volume(self, volume: int) -> None:
        """Decrease volume"""
        self.get_volume()
        if self.volume > 0:
            if (self.volume-volume) >= 0:
                self.media.audio_set_volume(
                    self.volume-volume
                )
                print(f"Volume: {self.media.audio_get_volume()}")
            else:
                self.media.audio_set_volume(0)
                self.volume = 0

    def get_length(self):
        """Get full track length: milliseconds"""
        return self.media.get_length()

    def get_time(self) -> int:
        """Get current media time: milliseconds"""
        return self.media.get_time()

    def remaining_time(self) -> int:
        """Get the remaining track time : seconds"""
        return (self.get_length()-self.get_time())/1000

    def seek_forward(self, time: int) -> None:
        """Seek forward"""
        self.media.set_time(
            self.get_time()+(time*1000)
        )

    def seek_backward(self, time: int) -> None:
        """Seek forward"""
        self.media.set_time(
            self.get_time()-(time*1000)
        )

    def get_state(self):
        """Get media state"""
        return self.media.get_state()

    def play(self):
        """Play"""
        self.media.play()

    def pause(self):
        """Pause"""
        self.media.pause()

    def stop(self):
        """Stop"""
        self.media.stop()


class MediaListPlayerControls():
    """Media List Player Controls"""

    def __init__(self, media_list_player: MediaListPlayer) -> None:
        self.media_list_player = media_list_player
        self.media_player = MediaPlayerControls(
            self.media_list_player.get_media_player())

    def play_index(self, index: int):
        """Plays item at certain index"""
        result = self.media_list_player.play_item_at_index(index)
        if result == 0:
            return self.media_player.get_song_url_from_player
        print("!*! Invalid Index/Song not found !*!")
        return None

    def next_song(self):
        """Plays next song in the list"""
        _next = self.media_list_player.next()
        if _next == 0:
            return self.media_player.get_song_url_from_player
        print("!*! Reached End of Media List !*!")
        return None

    def previous_song(self):
        """Plays previous song in the list"""
        _previous = self.media_list_player.previous()
        if _previous == 0:
            return self.media_player.get_song_url_from_player
        print("!*! Reached Start of Media List !*!")
        return None

    def play_start(self):
        """Play the first song in media list - start from beginning"""
        return self.media_list_player.play_item_at_index(0)

    def play(self):
        """Starts playing the media list"""
        return self.media_list_player.play()

    def stop(self):
        """Stops the playing media list"""
        return self.media_list_player.stop()

    def pause(self):
        """Pauses the playing media list"""
        return self.media_list_player.pause()


class StateControls():
    """Handler for State Controls"""

    def __init__(self, mplayer: MediaPlayerControls) -> None:
        self.mplayer_controls = mplayer
        self.input = None

    def manage(self, user_input: str):
        """Manages the player state"""

        self.input = user_input

        if self.input.lower() == 'p':

            if self.mplayer_controls.get_state() == State(4) or self.mplayer_controls.get_state() == State(5):  # 4 - Paused
                print("Resuming")
                self.mplayer_controls.play()

            elif self.mplayer_controls.get_state() == State(3):  # 3 - Playing (See State Enum definit.)
                print("Paused")
                self.mplayer_controls.pause()

        elif self.input.lower() == 's':
            print("Stopped")
            self.mplayer_controls.stop()

        elif (self.input.lower() == 'q' or self.input.lower() == 'e'):
            raise SystemExit("Bye! Bye!")


class VolumeControls():
    """Handler for Volume Controls"""

    def __init__(self, controls: MediaPlayerControls) -> None:
        self.controls = controls
        self.input = None

    def manage(self, user_input: str):
        """Manages volume increase/decrease"""

        self.input = user_input

        if self.input.startswith('+') and self.input.strip('+').isnumeric():
            vol = self.input.strip('+')
            if vol:
                self.controls.increase_volume(int(vol))
                # print(f"Volume: {self.controls.get_volume()}")

        elif self.input.startswith('-') and self.input.strip('-').isnumeric():
            vol = self.input.strip('-')
            if vol:
                self.controls.decrease_volume(int(vol))
                # print(f"Volume: {self.controls.get_volume()}")

        elif self.input.lower().startswith('v'):
            if self.input.lower().strip('v') and self.input.lower().strip('v').isnumeric():
                vol = self.input.lower().strip('v')
                if vol:
                    self.controls.set_volume(int(vol))
                    # print(f"Volume: {self.controls.get_volume()}")
            else:
                print(f"Volume: {self.controls.get_volume()}")


class MiscControls():
    """Handler for Misc Controls"""

    def __init__(self, controls: MediaPlayerControls) -> None:
        self.controls = controls
        self.input = None

    def manage(self, user_input: str):
        """Manages the Misc Options"""

        self.input = user_input

        if self.input.lower().startswith('lo'):
            # Pretty prints the available output devices.
            self.controls.display_devices()

        elif self.input.lower().startswith('co') and self.input.lower().strip('co').isnumeric():
            dev = self.input.lower().strip('co')
            if dev:
                self.controls.select_device(int(dev))

        elif self.input.lower().startswith('t'):
            print(f"Time Remaining {self.controls.remaining_time()} seconds")


class PlaybackControls():
    """Handler for Playback Controls"""

    def __init__(self, controls: MediaPlayerControls) -> None:
        self.controls = controls
        self.input = None

    def manage(self, user_input: str):
        """Manages the Playback Options"""

        self.input = user_input

        if self.input.startswith('>') and self.input.strip('>').isnumeric():
            seek = self.input.strip('>')
            if seek:
                self.controls.seek_forward(int(seek))
                print(
                    f"Forward: {seek}, RT - {self.controls.remaining_time()} seconds")

        elif self.input.startswith('<') and self.input.strip('<').isnumeric():
            seek = self.input.strip('<')
            if seek:
                self.controls.seek_backward(int(seek))
                print(
                    f"Backward: {seek}, RT - {self.controls.remaining_time()} seconds")


class MonitorState():
    """Handler for State Controls"""

    callback: Callable

    def __init__(self, controls: MediaPlayerControls) -> None:
        self.controls = controls
        _ = threading.Timer(5, self.manage)
        _.setDaemon(True)
        _.start()

    def manage(self):
        """Thread loop to monitor player status and update main song queue"""
        if self.controls.get_state() == State(3):
            self.callback('\033[31mPlayed\033[0m', self.controls.get_song_url_from_player)
        threading.Timer(15, self.manage).start()
        # if self.controls.get_state() == State(4):
        # elif self.controls.get_state() == State(6):
        # elif self.controls.get_state() == State(5):
